/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.Utilities;

import com.google.api.client.http.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.smartcardio.CardException;
import javax.smartcardio.CommandAPDU;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.*;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**
 * Utilities class.
 *
 * @author Ronald Koenig
 * @author André Szdzuy
 */
public class Utilities {
    private static final Log LOG = LogFactory.getLog(Utilities.class);

    /**
     * Convert an integer to a hexadecimal string.
     *
     * @param i integer to convert
     * @return String, containing hexadecimal representation
     */
    public static String toHexString(int i) {
        String result = Integer.toHexString(i);
        if ((result.length() % 2) != 0) {
            result = "0" + result;  // append leading zero if result length is uneven
        }
        return result;
    }

    /**
     * Utilities method to convert a byte array to a hexadecimal string.
     *
     * @param bytes bytes to convert
     * @return String, containing hexadecimal representation
     */
    public static String ByteArrayToHexString(byte[] bytes) {
        final char[] hexArray = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
        char[] hexChars = new char[bytes.length * 2];
        int v;
        for ( int j = 0; j < bytes.length; j++ ) {
            v = bytes[j] & 0xFF;
            hexChars[j * 2] = hexArray[v >>> 4];
            hexChars[j * 2 + 1] = hexArray[v & 0x0F];
        }
        return new String(hexChars);
    }

    /**
     * Utilities method to convert a byte array to a (better readable) hexadecimal string.
     *
     * @param bytes bytes to convert
     * @return String, containing (better readable) hexadecimal representation
     */
    public static String ByteArrayToSpacedHexString(byte[] bytes) {
        final char[] hexArray = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
        char[] hexChars = new char[bytes.length * 3];
        int v;
        for ( int j = 0; j < bytes.length; j++ ) {
            v = bytes[j] & 0xFF;
            hexChars[j * 3] = hexArray[v >>> 4];
            hexChars[j * 3 + 1] = hexArray[v & 0x0F];
            hexChars[j * 3 + 2] = ' ';
        }
        return new String(hexChars);
    }

    /**
     * Utilities method to convert a hexadecimal string to a byte string.
     *
     * <p>Behavior with input strings containing non-hexadecimal characters is undefined.
     *
     * @param s String containing hexadecimal characters to convert
     * @return Byte array generated from input
     */
    public static byte[] HexStringToByteArray(String s) {
        s=s.replaceAll("\\s","");
        int len = s.length();
        byte[] data = new byte[len / 2];
        for (int i = 0; i < len; i += 2) {
            data[i / 2] = (byte) ((Character.digit(s.charAt(i), 16) << 4)
                    + Character.digit(s.charAt(i+1), 16));
        }
        return data;
    }

    /**
     * Utilities method to read text content from an input stream. Use ISO8859-15 as default encoding.
     *
     * @param is input stream
     * @return text content
     * @throws IOException if an error occurred
     */
    public static String readTextFromStream(InputStream is) throws IOException {
        return readTextFromStream(is, "ISO-8859-1");
    }

    /**
     * Utilities method to read text content from an input stream.
     *
     * @param is input stream
     * @param encoding encoding to use
     * @return text content
     * @throws IOException if an error occurred
     */
    public static String readTextFromStream(InputStream is, String encoding) throws IOException {
        Writer writer = new StringWriter();
        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(is, encoding));
            String line = reader.readLine();
            while (line != null) {
                writer.write(line);
                line = reader.readLine();
            }
        } finally {
            try {
                is.close();
            } catch (Exception e) {
                LOG.error("Exception when closing input stream", e);
            }
        }

        return writer.toString();
    }

    /**
     * Utilities method to generate a X509Certificate from the given data.
     *
     * @param data certificate data
     * @return an X509Certificate instance
     * @throws CertificateException if an error occurred
     */
    public static X509Certificate getX509Certificate(byte[] data) throws CertificateException {
        X509Certificate x509Cert = null;
        ByteArrayInputStream certInputStream = new ByteArrayInputStream(data);
        CertificateFactory certFactory = CertificateFactory.getInstance("X509");
        if (certFactory != null) {
            x509Cert = (X509Certificate) certFactory.generateCertificate(certInputStream);
        }

        return x509Cert;
    }

    /**
     * Unzip text content. Use ISO8859-15 as default encoding.
     *
     * @param data the zipped data
     * @return the unzipped text content
     * @throws IOException if an error occurred
     */
    public static String unzip(byte[] data) throws IOException {
        return unzip(data, "ISO8859-15");
    }

    /**
     * Unzip text content.
     *
     * @param data the zipped data
     * @param encoding encoding to use
     * @return the unzipped text content
     * @throws IOException if an error occurred
     */
    public static String unzip(byte[] data, String encoding) throws IOException {
        return unzip(new ByteArrayInputStream(data), encoding);
    }

    /**
     * Unzip text content. Use ISO8859-15 as default encoding.
     *
     * @param data the zipped data
     * @param offset offset
     * @param length length
     * @return the unzipped text content
     * @throws IOException if an error occurred
     */
    public static String unzip(byte[] data, int offset, int length) throws IOException {
        return unzip(new ByteArrayInputStream(data, offset, length), "ISO8859-15");
    }

    private static String unzip(ByteArrayInputStream bis, String encoding) throws IOException {
        ByteArrayOutputStream bos = null;
        GZIPInputStream zis = null;
        try {
            zis = new GZIPInputStream(bis);
            bos = new ByteArrayOutputStream();
            byte[] buffer = new byte[1024];
            int count;
            while ((count = zis.read(buffer)) != -1) {
                bos.write(buffer, 0, count);
            }
            return bos.toString(encoding);
        } finally {
            try {
                if (bos != null) {
                    bos.close();
                }
            } catch (IOException e) {
                LOG.info("Cannot close bos");
            }
            try {
                if (bis != null) {
                    bis.close();
                }
            } catch (IOException e) {
                LOG.info("Cannot close bis");
            }
            try {
                if (zis != null) {
                    zis.close();
                }
            } catch (IOException e) {
                LOG.info("Cannot close zis");
            }
        }
    }

    /**
     * Zip text content. Use ISO8859-15 as default encoding.
     *
     * @param content the text content
     * @return the zipped data
     * @throws IOException if an error occurred
     */
    public static byte[] zip(String content) throws IOException {
        return zip(content, "ISO8859-15");
    }

    /**
     * Zip text content.
     *
     * @param content the text content
     * @param encoding encoding to use
     * @return the zipped data
     * @throws IOException if an error occurred
     */
    public static byte[] zip(String content, String encoding) throws IOException {
        ByteArrayInputStream bis = new ByteArrayInputStream(content.getBytes(encoding));
        ByteArrayOutputStream bos = null;
        GZIPOutputStream zos = null;
        try {
            bos = new ByteArrayOutputStream();
            zos = new GZIPOutputStream(bos);
            byte[] buffer = new byte[1024];
            int count;
            while ((count = bis.read(buffer)) != -1) {
                zos.write(buffer, 0, count);
            }
            zos.finish();
            return bos.toByteArray();
        } finally {
            try {
                if (bos != null) {
                    bos.close();
                }
            } catch (IOException e) {
                LOG.info("Cannot close bos");
            }
            try {
                if (bis != null) {
                    bis.close();
                }
            } catch (IOException e) {
                LOG.info("Cannot close bis");
            }
            try {
                if (zos != null) {
                    zos.close();
                }
            } catch (IOException e) {
                LOG.info("Cannot close zos");
            }
        }
    }

    /**
     * Transform the given XML content to a DOM and return the root element of the DOM.
     *
     * @param xmlDoc String with XML content
     *
     * @return root element of the DOM
     * @throws IOException if an error occurred
     * @throws SAXException if an error occurred
     * @throws ParserConfigurationException if an error occurred
     */
    public static Element getDOMRootElement(String xmlDoc) throws IOException, SAXException, ParserConfigurationException {
        DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
        Document doc = dBuilder.parse(new InputSource(new StringReader(xmlDoc)));
        doc.getDocumentElement().normalize();
        return doc.getDocumentElement();
    }

    /**
     * Display the given XML content (all in one line) in a better readable form with newlines and
     * indentations.
     *
     * @param xmlDoc String with XML content
     *
     * @return better readable form
     * @throws IOException if an error occurred
     * @throws SAXException if an error occurred
     * @throws ParserConfigurationException if an error occurred
     */
    public static String displayXmlElements(String xmlDoc) throws IOException, SAXException, ParserConfigurationException {
        StringBuffer buf = new StringBuffer();
        processXmlElement(getDOMRootElement(xmlDoc), buf, "");
        return buf.toString();
    }

    private static void processXmlElement(Element e, StringBuffer buf, String indent) {
        NodeList nList = e.getChildNodes();
        List<Element> elements = new ArrayList<Element>();

        // first select elements from node list
        for (int i = 0; i < nList.getLength(); i++) {
            Node nNode = nList.item(i);
            if (nNode.getNodeType() == Node.ELEMENT_NODE) {
                elements.add((Element) nNode);
            }
        }

        // then process elements
        String tagName = e.getTagName().trim();
        if (elements.size() == 0) {   // end of recursion
            buf.append(indent + tagName + ": ");
            buf.append(e.getTextContent() + "\n");
        } else {
            buf.append("\n" + indent + tagName + "\n" + indent);
            for (int i = 0; i < tagName.length(); i++) {
                buf.append("=");
            }
            buf.append("\n");
            for (Iterator<Element> it = elements.iterator(); it.hasNext(); ) {
                Element child = it.next();
                processXmlElement(child, buf, indent + "   ");
            }
        }
    }

    /**
     * Remove <count> lines from the given <text>.
     */
    public static String removeLines(String text, int count) {
        String s = text;
        for (int i = 0; i < count; i++) {
            int pos = s.indexOf("\n");
            s = s.substring(pos + 1);
        }
        return s;
    }

    /**
     * Write details of a CommandAPDU to the log.
     *
     * @param commandAPDU CommandAPDU
     */
    public static void logCommandAPDU(CommandAPDU commandAPDU) {
        LOG.info("transmit: " + commandAPDU.toString()
                + ", CLA=" + Utilities.toHexString(commandAPDU.getCLA())
                + ", INS=" + Utilities.toHexString(commandAPDU.getINS())
                + ", P1=" + Utilities.toHexString(commandAPDU.getP1())
                + ", P2=" + Utilities.toHexString(commandAPDU.getP2()));
        LOG.info("      Data=" + Utilities.ByteArrayToSpacedHexString(commandAPDU.getData()));
        LOG.info("      complete APDU=" + Utilities.ByteArrayToSpacedHexString(commandAPDU.getBytes()));
    }

    /**
     * Send a HTTP POST request with the given binary content in the request body.
     * Also waits for the response and extracts possible binary content from the response body.
     *
     * @param httpRequestFactory HttpRequestFactory instance to use for request creation
     * @param url URL to send the request to
     * @param content binary content
     * @return binary content of the response
     * @throws IOException if an error occurred
     */
    public static byte[] sendBinaryHttpRequest(HttpRequestFactory httpRequestFactory, String url, byte[] content) throws IOException {
        return sendBinaryHttpRequest(httpRequestFactory, url, content, null);
    }

    /**
     * Send a HTTP POST request with the given binary content in the request body.
     * Also waits for the response and extracts possible binary content from the response body.
     *
     * @param httpRequestFactory HttpRequestFactory instance to use for request creation
     * @param url URL to send the request to
     * @param content binary content
     * @param headers optional HTTP headers
     * @return binary content of the response
     * @throws IOException if an error occurred
     */
    public static byte[] sendBinaryHttpRequest(HttpRequestFactory httpRequestFactory, String url, byte[] content, HttpHeaders headers) throws IOException {
        GenericUrl genericUrl = new GenericUrl(url);
        return sendBinaryHttpRequest(httpRequestFactory, genericUrl, content, headers);
    }

    /**
     * Send a HTTP POST request with the given binary content in the request body.
     * Also waits for the response and extracts possible binary content from the response body.
     *
     * @param httpRequestFactory HttpRequestFactory instance to use for request creation
     * @param genericUrl URL to send the request to
     * @param content binary content
     * @param headers optional HTTP headers
     * @return binary content of the response
     * @throws IOException if an error occurred
     */
    public static byte[] sendBinaryHttpRequest(HttpRequestFactory httpRequestFactory, GenericUrl genericUrl, byte[] content, HttpHeaders headers) throws IOException {
        LOG.info("sendBinaryHttpRequest()");
        LOG.info("     URL: " + genericUrl.toString());
        LOG.info("     send: " + Utilities.ByteArrayToSpacedHexString(content));
        HttpRequest httpRequest = httpRequestFactory.buildPostRequest(
                genericUrl, new ByteArrayContent("application/octet-stream", content));
        httpRequest.setConnectTimeout(1000);
        if (headers != null) {
            httpRequest.setHeaders(headers);
        }
        HttpResponse httpResponse = httpRequest.execute();  // throws exception if response has any HTTP error code
        try {
            InputStream is = httpResponse.getContent();
    //        if (is != null && is.available() > 0) {
            if (is != null) {   // don't check is.available()>0 because on Android it always returns 0
                byte[] tmpBuffer = new byte[1024];
                int length = is.read(tmpBuffer);
                byte[] responseContent = new byte[length];
                System.arraycopy(tmpBuffer, 0, responseContent, 0, length);
                LOG.info("     received: " + Utilities.ByteArrayToSpacedHexString(responseContent));
                return responseContent;
            } else {
                LOG.info("     received: <empty>");
                return null;
            }
        } finally {
            httpResponse.disconnect();
        }
    }

    /**
     * Adjust channel. The given CLA byte is "enriched" with the given channel number.
     *
     * @param channelNumber channel number
     * @param cla CLA byte
     * @return adjusted CLA byte
     * @throws CardException if an error occurred
     */
    public static byte adjustChannel(int channelNumber, byte cla) throws CardException {
        int newCla;
        if (channelNumber < 4) {
            newCla = (cla & 0xfc) | channelNumber;
        } else if (channelNumber < 16) {
            newCla = cla & 0x80; //keep bit 7
            newCla |= 0x40; //set bit 6
            newCla |= (cla & 0x1c) << 1; //shift bit 4-2 by 1
            newCla |= channelNumber;
        } else {
            throw new CardException("channel number too big: " + channelNumber);
        }
        return (byte) newCla;
    }

}
