/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.card2server;

import com.google.api.client.http.GenericUrl;
import com.google.api.client.http.HttpRequestFactory;
import com.google.api.client.http.HttpTransport;
import com.google.api.client.http.javanet.NetHttpTransport;
import de.gematik.security.Utilities.Utilities;
import de.gematik.security.impl.HealthcardContext;
import de.gematik.security.iso7816.Constants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.smartcardio.CardException;
import javax.smartcardio.CommandAPDU;
import javax.smartcardio.ResponseAPDU;
import java.io.IOException;

/**
 * Represents the EGK layer in the Card2Server descriptions. This layer receives apdu commands from the
 * "VSDD Steuersoftware" and executes them on the EGK card.
 *
 * @author André Szdzuy
 */
public class Client {
    private static final Log LOG = LogFactory.getLog(Client.class);

    private HealthcardContext healthcardContext = null;
    private String serverURL = null;

    public Client(HealthcardContext healthcardContext, String serverURL) {
        this.healthcardContext = healthcardContext;
        this.serverURL = serverURL;
    }

    public void updateVSDM() throws CardException {
        try {
            HttpTransport httpTransport = new NetHttpTransport();
            HttpRequestFactory httpRequestFactory = httpTransport.createRequestFactory();

            byte[] iccsnEGK = healthcardContext.readICCSN();
            LOG.info("iccsnEGK: " + Utilities.ByteArrayToSpacedHexString(iccsnEGK));

            // this removes the PACE SecureMessaging because it conflicts with the server SecureMessaging
            healthcardContext.disableSecureMessaging();

            // GET CHALLENGE -> card
            byte[] apduGetChallenge = sendStartUpdateRequest(httpRequestFactory, serverURL + "/startUpdate", iccsnEGK,
                    healthcardContext.getChannel().getChannelNumber());

            ResponseAPDU responseAPDU = healthcardContext.getChannel().transmit(new CommandAPDU(apduGetChallenge));
            healthcardContext.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
            byte[] rnd1 = responseAPDU.getData();
            LOG.info("rnd1: " + Utilities.ByteArrayToSpacedHexString(rnd1));

            // MSE SET -> card (select key for MUTUAL AUTHENTICATE)
            String keyRef = Utilities.toHexString(0x00 | 0x19);   // location = 0x00, identifier = 0x19
            String algId = "54";    // aesSessionkey4SM (MF/ SK.VSD.AES256)
            healthcardContext.manageSE("83 01", keyRef, algId, 0x81, 0xA4);

            // MUTUAL AUTHENTICATE -> card
            byte[] apduMutualAuthenticate =
                    Utilities.sendBinaryHttpRequest(httpRequestFactory, serverURL + "/responseGetChallenge", rnd1);

            responseAPDU = healthcardContext.getChannel().transmit(new CommandAPDU(apduMutualAuthenticate));
            healthcardContext.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
            byte[] maOutput = responseAPDU.getData();

            //SELECT DF.HCA -> card
            byte[] apduSelectDfHCA =
                    Utilities.sendBinaryHttpRequest(httpRequestFactory, serverURL + "/responseMutualAuthenticate", maOutput);

            responseAPDU = healthcardContext.getChannel().transmit(new CommandAPDU(apduSelectDfHCA));
            healthcardContext.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);

            //SELECT EF.StatusVD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseSelectDfHCA", responseAPDU);

            //UPDATE BINARY EF.StatusVD: status=1 -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseSelectEfStatusVD", responseAPDU);

            //SELECT EF.PD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseUpdateEfStatusVD", responseAPDU);

            //UPDATE BINARY EF.PD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseSelectEfPD", responseAPDU);

            //SELECT EF.VD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseUpdateEfPD", responseAPDU);

            //UPDATE BINARY EF.VD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseSelectEfVD", responseAPDU);

            //SELECT EF.GVD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseUpdateEfVD", responseAPDU);

            //UPDATE BINARY EF.GVD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseSelectEfGVD", responseAPDU);

            //SELECT EF.StatusVD -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/responseSelectDfHCA", responseAPDU);

            //UPDATE BINARY EF.StatusVD: status=0 -> card
            responseAPDU = processNextCommand(httpRequestFactory, "/finish", responseAPDU);
        } catch (Exception e) {
            LOG.error(e.getMessage());
            throw new CardException(e);
        } finally {
            // this restores the PACE SecureMessaging if there was any before entering this method
            healthcardContext.enableSecureMessaging();
        }
    }

    private ResponseAPDU processNextCommand(HttpRequestFactory httpRequestFactory, String serverURI, ResponseAPDU previousResponse) throws CardException, IOException {
        byte[] commandAPDU =
                Utilities.sendBinaryHttpRequest(httpRequestFactory, serverURL + serverURI, previousResponse.getBytes());

        ResponseAPDU responseAPDU = healthcardContext.getChannel().transmit(new CommandAPDU(commandAPDU));
        healthcardContext.assertResponseStatus(responseAPDU, Constants.APDU_RESPONSE_OK);
        return responseAPDU;
    }

    private byte[] sendStartUpdateRequest(HttpRequestFactory httpRequestFactory, String url, byte[] content, int channelNumber) throws IOException {
        GenericUrl genericUrl = new GenericUrl(url);
        genericUrl.set("cardChannel", channelNumber);
        return Utilities.sendBinaryHttpRequest(httpRequestFactory, genericUrl, content, null);
    }

}
