/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.test;

import de.gematik.security.Utilities.CardFileSystem;
import de.gematik.security.Utilities.Utilities;
import de.gematik.security.impl.HealthcardContext;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;
import de.gematik.security.iso9564.PinBlockException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.junit.Test;

import javax.smartcardio.*;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ShortBuffer;
import java.util.List;

/**
 * @author André Szdzuy
 */
public class Card2CardTest extends NfcBaseTest {
    private static final Log LOG = LogFactory.getLog(Card2CardTest.class);

    @Test
    public void connectSecondCard() throws CardException, Iso7816UnsuccessfulExecutionException, PinBlockException, IOException {
        LOG.info("connectSecondCard()");

        Card nfcCard2 = null;
        HealthcardContext healthcardContext2 = null;
        CardFileSystem cfs2 = null;

        try {
            while (true) {
                CardTerminals cardTerminals = TerminalFactory.getDefault().terminals();
                LOG.info("Please insert second card!");
                cardTerminals.waitForChange();
                List<CardTerminal> list = cardTerminals.list(CardTerminals.State.ALL);
                for (int i = 0; i < list.size(); i++) {
                    CardTerminal item = list.get(i);
                    if (item != null && !cardTerminalName.equals(item.getName()) && item.isCardPresent()) {
                        LOG.info("Connect to card terminal " + item.getName());
                        nfcCard2 = item.connect("*");
                        break;
                    }
                }

                if (nfcCard2 != null) {
                    LOG.info("Open secure channel");
                    healthcardContext2 = new HealthcardContext(nfcCard2.openLogicalChannel(), true);
                    if (!healthcardContext2.isSMCB()) {
                        // strange: without this sleep secureChannel() fails with code 0x6a80 if second card is
                        // inserted into Reiner SCT (if second card is inserted into SCR33x it works without sleep)
                        try {
                            Thread.sleep(500);
                        } catch (InterruptedException e) {
                            LOG.error(e);
                        }
                        healthcardContext2.secureChannel("123123");
                        LOG.info("Pace established");
                    }

                    if (healthcardContext2.isEGK()) {
                        cfs2 = CardFileSystem.getEGKInstance();
                    } else if (healthcardContext2.isHBA()) {
                        cfs2 = CardFileSystem.getHBAInstance();
                    } else if (healthcardContext2.isSMCB()) {
                        cfs2 = CardFileSystem.getSMCBInstance();
                    } else {
                        throw new CardException("Inserted card is neither EGK, HBA nor SMC-B.");
                    }

                    break;
                }
            }

            LOG.info("Both cards are connected now.");
            LOG.info("Demonstrate access to first card by checking its type: " + healthcardContext.getType());
            LOG.info("Demonstrate access to second card by checking its type: " + healthcardContext2.getType());

            HealthcardContext egk;
            HealthcardContext hba;

            if (healthcardContext.isEGK()) {
                egk = healthcardContext;
                hba = healthcardContext2;
            } else {
                egk = healthcardContext2;
                hba = healthcardContext;
            }

            // test Card2Card authentication:

            // Use Case 1: Freischalten einer EGK durch einen HBA, z.B. zum Zugriff auf die Notfalldaten (NFD)
            // precondition: PIN.CH/PIN.SMC verified for HBA/SMC-B
            hba.verify("123456");
            egk.authenticateOtherCard(hba);

            // Use Case 2: Echtheitsprüfung einer EGK durch einen HBA
            hba.authenticateOtherCard(egk);

            // test access to NFD in order to verify the success of Card2Card authentication:

            //SELECT DF.HCA
            egk.selectADF("D27600000102");

            //SELECT DF.HCA/DF.NFD
            egk.selectADF("D27600014407");

            //SELECT EF.NFD
            egk.selectEF("D010");

            if (egk.isExtendedLengthApduSupported()) {
                egk.setMaxLength(4096); // readBinary funktioniert nur bis LE=4096, LE=8192 und größer liefert den Error 6F00
            }

            // Write NFD data
            InputStream is = new FileInputStream("C:/Dateien/NFD_Daten_BÄK_MP_2.xml");
            String xmlNFD = Utilities.readTextFromStream(is, "UTF-8");
            byte[] zippedNFD = Utilities.zip(xmlNFD, "UTF-8");
            byte[] lengthBytes = Utilities.HexStringToByteArray(Utilities.toHexString(zippedNFD.length));

            byte[] data = new byte[zippedNFD.length + 2];
            System.arraycopy(lengthBytes, 0, data, 2 - lengthBytes.length, lengthBytes.length);
            System.arraycopy(zippedNFD, 0, data, 2, zippedNFD.length);

            egk.update(data);

            // Read length
            ShortBuffer shortBuffer = ByteBuffer.wrap(egk.read(0, 2)).asShortBuffer();

            // Read data
            data = egk.read(2, shortBuffer.get(0));
            String result = Utilities.unzip(data);
            LOG.info(result);

        } catch (Exception e) {
            LOG.error(e.getMessage());
            throw e;
        } finally {
            if (nfcCard2 != null) {
                nfcCard2.disconnect(true);
            }
            if (healthcardContext2 != null) {
                healthcardContext2.close();
            }

        }
    }

}
