/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.test;

import de.gematik.security.Utilities.Utilities;
import de.gematik.security.iso7816.FCP;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;
import de.gematik.security.iso9564.PinBlock;
import de.gematik.security.iso9564.PinBlockException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.junit.Before;
import org.junit.Test;

import javax.smartcardio.CardException;
import java.nio.ByteBuffer;
import java.nio.ShortBuffer;

import static org.junit.Assert.*;

/**
 * @author André Szdzuy
 */
public class Iso7816CommandsTest extends NfcBaseTest {
    private static final Log LOG = LogFactory.getLog(Iso7816CommandsTest.class);

    @Before
    public void setUp() throws Exception {
        assertNotNull("healthcardContext is null", healthcardContext);

        // select root file before each test
        healthcardContext.selectADF();
    }

    @Test
    public void selectADF() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADF()");
        assertNotNull("healthcardContext is null", healthcardContext);
        healthcardContext.selectADF();
        // no assert necessary here because result is already checked in implementation
    }

    @Test
    public void selectADFWithNextOccurrence() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADFWithNextOccurrence()");
        assertNotNull("healthcardContext is null", healthcardContext);
        healthcardContext.selectADF(true);
        // no assert necessary here because result is already checked in implementation
    }

    @Test
    public void selectADFWithApplicationIdentifier() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADFWithApplicationIdentifier()");
        assertNotNull("healthcardContext is null", healthcardContext);
        healthcardContext.selectADF("A000000167455349474E");
        // no assert necessary here because result is already checked in implementation
    }

    @Test
    public void selectADFWithApplicationIdentifierAndNextOccurrence() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADFWithApplicationIdentifierAndNextOccurrence()");
        assertNotNull("healthcardContext is null", healthcardContext);
        healthcardContext.selectADF("A000000167455349474E", true);
        // no assert necessary here because result is already checked in implementation
    }

    @Test
    public void selectDF() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectDF()");
        assertNotNull("healthcardContext is null", healthcardContext);
        healthcardContext.selectADF("A000000167455349474E");   // necessary pre-step
        healthcardContext.selectDF();
        // no assert necessary here because result is already checked in implementation
    }

    @Test
    public void selectDFWithFileIdentifier() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectDFWithFileIdentifier()");
        LOG.info("Not applicable for eGK because there exist no DFs with a fileIdentifier.");
    }

    @Test
    public void selectDFWithFileIdentifierAndNextOccurrence() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectDFWithFileIdentifierAndNextOccurrence()");
        LOG.info("Not applicable for eGK because there exist no DFs with a fileIdentifier.");
    }

    @Test
    public void selectEF() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectEF()");
        assertNotNull("healthcardContext is null", healthcardContext);
        healthcardContext.selectADF("A000000167455349474E");   // necessary pre-step
        healthcardContext.selectEF("C500");
        // no assert necessary here because result is already checked in implementation
    }

    @Test
    public void selectADFAndReturnFCP() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADFAndReturnFCP()");
        assertNotNull("healthcardContext is null", healthcardContext);

        FCP fcp = healthcardContext.selectADFAndReturnFCP();
        String result = Utilities.ByteArrayToSpacedHexString(fcp.getDfName());
        LOG.info("Root application identifier: " + result);

        String expectedResult;
        if (healthcardContext.isEGK()) {
            expectedResult = "D2 76 00 01 44 80 00";
        } else if (healthcardContext.isHBA()) {
            expectedResult = "D2 76 00 01 46 01";
        } else {    // SMC-B
            expectedResult = "D2 76 00 01 46 06";
        }
        assertEquals("Unexpected result", expectedResult, result.trim());
    }

    @Test
    public void selectADFAndReturnFCPWithNextOccurrence() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADFAndReturnFCPWithNextOccurrence()");
        assertNotNull("healthcardContext is null", healthcardContext);

        FCP fcp = healthcardContext.selectADFAndReturnFCP(true);
        String result = Utilities.ByteArrayToSpacedHexString(fcp.getDfName());

        String expectedResult;
        if (healthcardContext.isEGK()) {
            expectedResult = "E8 28 BD 08 0F A0 00 00 01 67 45 53 49 47 4E";
        } else if (healthcardContext.isHBA()) {
            expectedResult = "D2 76 00 01 46 03";
        } else {    // SMC-B
            expectedResult = "A0 00 00 01 67 45 53 49 47 4E";
        }
        assertEquals("Unexpected result", expectedResult, result.trim());
    }

    @Test
    public void selectADFAndReturnFCPWithApplicationIdentifier() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADFAndReturnFCPWithApplicationIdentifier()");
        assertNotNull("healthcardContext is null", healthcardContext);

        FCP fcp = healthcardContext.selectADFAndReturnFCP("A000000167455349474E");
        String result = Utilities.ByteArrayToSpacedHexString(fcp.getDfName());
        assertEquals("Unexpected result", "A0 00 00 01 67 45 53 49 47 4E", result.trim());
    }

    @Test
    public void selectADFAndReturnFCPWithApplicationIdentifierAndNextOccurrence() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectADFAndReturnFCPWithApplicationIdentifierAndNextOccurrence()");
        assertNotNull("healthcardContext is null", healthcardContext);

        FCP fcp = healthcardContext.selectADFAndReturnFCP("A000000167455349474E", true);
        String result = Utilities.ByteArrayToSpacedHexString(fcp.getDfName());
        assertEquals("Unexpected result", "A0 00 00 01 67 45 53 49 47 4E", result.trim());
    }

    @Test
    public void selectDFAndReturnFCP() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectDFAndReturnFCP()");
        assertNotNull("healthcardContext is null", healthcardContext);

        healthcardContext.selectADF("A000000167455349474E");   // necessary pre-step
        FCP fcp = healthcardContext.selectDFAndReturnFCP();
        String result = Utilities.ByteArrayToSpacedHexString(fcp.getDfName());

        String expectedResult;
        if (healthcardContext.isEGK()) {
            expectedResult = "D2 76 00 01 44 80 00";
        } else if (healthcardContext.isHBA()) {
            expectedResult = "D2 76 00 01 46 01";
        } else {    // SMC-B
            expectedResult = "D2 76 00 01 46 06";
        }
        assertEquals("Unexpected result", expectedResult, result.trim());
    }

    @Test
    public void selectDFAndReturnFCPWithFileIdentifier() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectDFAndReturnFCPWithFileIdentifier()");
        LOG.info("Not applicable for eGK because there exist no DFs with a fileIdentifier.");
    }

    @Test
    public void selectDFAndReturnFCPWithFileIdentifierAndNextOccurrence() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectDFAndReturnFCPWithFileIdentifierAndNextOccurrence()");
        LOG.info("Not applicable for eGK because there exist no DFs with a fileIdentifier.");
    }

    @Test
    public void selectEFAndReturnFCPWithFileIdentifier() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("selectEFAndReturnFCPWithFileIdentifier()");
        assertNotNull("healthcardContext is null", healthcardContext);

        healthcardContext.selectADF("A000000167455349474E");   // necessary pre-step
        FCP fcp = healthcardContext.selectEFAndReturnFCP("C500");
        String result = Utilities.ByteArrayToSpacedHexString(fcp.getFileIdentifier());
        assertEquals("Unexpected result", "C5 00", result.trim());
    }

    @Test
    public void read() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("read()");
        assertNotNull("healthcardContext is null", healthcardContext);

        healthcardContext.selectADF("A000000167455349474E");   // necessary pre-step
        healthcardContext.selectEF("C500");   // necessary pre-step
        byte[] result = healthcardContext.read();
        assertNotNull("read() returned null", result);

        // to check the length of the data is not useful here because it is different for each card
    }

    @Test
    public void readWithParameters() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("readWithParameters()");
        assertNotNull("healthcardContext is null", healthcardContext);

        healthcardContext.selectADF("A000000167455349474E");   // necessary pre-step
        healthcardContext.selectEF("C500");   // necessary pre-step
        byte[] bytes = healthcardContext.read(0, 2);
        assertNotNull("read() returned null", bytes);
        ShortBuffer shortBuffer = ByteBuffer.wrap(bytes).asShortBuffer();
        assertEquals("Unexpected result", 12418, shortBuffer.get(0));
    }

    @Test
    public void readRecord() throws Iso7816UnsuccessfulExecutionException, CardException {
        LOG.info("readRecord()");
        assertNotNull("healthcardContext is null", healthcardContext);

        healthcardContext.selectEF("2F00");   // necessary pre-step
        byte[] result = healthcardContext.readRecord(1);
        assertNotNull("readRecord() returned null", result);

        long expectedResult;
        if (healthcardContext.isEGK()) {
            expectedResult = 11;
        } else if (healthcardContext.isHBA()) {
            expectedResult = 10;
        } else {    // SMC-B
            expectedResult = 10;
        }
        assertEquals("Unexpected result", expectedResult, result.length);
    }

    @Test
    public void verify() throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
        LOG.info("verify()");
        assertNotNull("healthcardContext is null", healthcardContext);

        healthcardContext.verify("123456");
    }

    @Test
    public void verifyWithPinBlock() throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
        LOG.info("verifyWithPinBlock()");
        assertNotNull("healthcardContext is null", healthcardContext);

        if (healthcardContext.isEGK()) {
            healthcardContext.verify(2, new PinBlock("123456"));
        } else {    // HBA or SMC-B
            healthcardContext.verify(1, new PinBlock("123456"));
        }
    }

    @Test
    public void verifyWithWrongData() throws CardException, PinBlockException {
        LOG.info("verifyWithWrongData()");
        assertNotNull("healthcardContext is null", healthcardContext);

        try {
            healthcardContext.verify("5555");
            fail("verify() accepted wrong PIN");
        } catch (Iso7816UnsuccessfulExecutionException e) {
            assertEquals("verify() failed with unexpected error code", 0x63c2, e.getSW());
        }
    }

    // do not run the unlock tests automatically because the card becomes unusable after 10 unlocks!
    // this is not reversible!

//    @Test
//    public void unlockWithPUK() throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
//        LOG.info("unlockWithPUK()");
//        assertNotNull("healthcardContext is null", healthcardContext);
//
//        // first call verify() with wrong PIN several times
//        for (int i = 0; i < 5 ; i++) {
//            try {
//                healthcardContext.verify("5555");
//                fail("verify() accepted wrong PIN");
//            } catch (Iso7816UnsuccessfulExecutionException e) {
//                LOG.info("Result of verify(): " + e.getMessage());
//            }
//        }
//
//        // then call verify() with right PIN which should fail
//        try {
//            healthcardContext.verify("123456");
//            fail("verify() accepted PIN although it should be blocked");
//        } catch (Iso7816UnsuccessfulExecutionException e) {
//            assertEquals("verify() failed with unexpected error code", 0x6983, e.getSW());
//        }
//
//        // then unlock
//        healthcardContext.unlock("12345678", null);
//
//        // finally call verify() with right PIN again
//        healthcardContext.verify("123456");
//    }
//
//    @Test
//    public void unlockWithPUKandNewPIN() throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException {
//        LOG.info("unlockWithPUKandNewPIN()");
//        assertNotNull("healthcardContext is null", healthcardContext);
//
//        // first call verify() with wrong PIN several times
//        for (int i = 0; i < 5 ; i++) {
//            try {
//                healthcardContext.verify("5555");
//                fail("verify() accepted wrong PIN");
//            } catch (Iso7816UnsuccessfulExecutionException e) {
//                LOG.info("Result of verify(): " + e.getMessage());
//            }
//        }
//
//        // then call verify() with right PIN which should fail
//        try {
//            healthcardContext.verify("123456");
//            fail("verify() accepted PIN although it should be blocked");
//        } catch (Iso7816UnsuccessfulExecutionException e) {
//            assertEquals("verify() failed with unexpected error code", 0x6983, e.getSW());
//        }
//
//        // then unlock
//        healthcardContext.unlock("12345678", "999999");
//
//        // finally call verify() with the new PIN
//        healthcardContext.verify("999999");
//    }

}
